/* 
 * Copyright (C) 1999-2001 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "jftpgw.h"
#include "fw_auth_cmds.h"
#include <sys/mman.h>

#define ANON_USERS " anonymous ftp "

extern int timeout;
extern struct loginfo li;
extern struct hostent_list* hostcache;
extern struct conn_info_st conn_info;

#ifdef SFTP_SUPPORT
static int sftp_login(struct clientinfo*);
#endif

static int login_setforward(struct clientinfo*);
static int login_connect(struct clientinfo*);
static int login_mayconnect(struct clientinfo*);
static int login_init_connection(struct clientinfo*);
static int login_connected_setup(struct clientinfo*);

static int login_auth(struct clientinfo*);
static int login_readwelcome(struct clientinfo*);
static int login_sendauth_user(struct clientinfo*);
static int login_sendauth_pass(struct clientinfo*);
static int login_successfulp(struct clientinfo*);
static int login_loggedin_setup(struct clientinfo*);

static int login_failed(struct clientinfo*);


int handle_login(struct clientinfo* clntinfo) {
	struct cmdhandlerstruct *cmdhandler;
	char *buffer = 0;
	int ss, cs;
	int i, expected;

	conn_info.li = &li;
	conn_info.clntinfo = clntinfo;
	ss = clntinfo->serversocket = -1;
	cs = clntinfo->clientsocket;

	cmdhandler = &login_auth_funcs
				[ config_get_ioption("loginstyle", 0) ][0];


	expected = QUITFUNC + 1;    /* skip reset and quit function */
	while (1) {
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			return -1;
		}
		if (buffer) {
			free(buffer);
			buffer = 0;
		}

		buffer = readline(cs);
		li.cmd = (char*) 0;

		if (buffer) {
			li.cmd = buffer;
			free(li.method);
			i = 0;
			li.method = quotstrtok(li.cmd, WHITESPACES, &i);

			if (buffer[0] == '\0') {
				/* empty line. Prevent logging of the
				 * command */
				free(buffer);
				buffer = 0;
				continue;
			}
			/* log the command */
			log_cmd(&li);
			free(li.method); li.method = (char*) 0;
			li.respcode = 0;
			li.transferred = 0;
		}

		if (!buffer) {
			if (timeout) {
				log(2, "Timeout in %s line %d\n", __FILE__
						,__LINE__);
				err_time_readline(cs);
			} else {
				err_readline(cs);
			}
			return -1;
		}

		log(8, "Expecting %s", cmdhandler[expected].cmd);
		log(8, "Got: %s", buffer);

		/* check for QUIT */
		if (checkbegin(buffer, "QUIT")) {
			int ret = (cmdhandler[QUITFUNC].func)
						(buffer, &conn_info);
			ret = (cmdhandler[RESETFUNC].func)
						(buffer, &conn_info);
			free(buffer);
			/* return 1 to prepend the
			 * proxy enters handle_cmds
			 * */
			return 1;
		}
		if (cmdhandler[expected].cmd 
			&& checkbegin(buffer, cmdhandler[expected].cmd)) {

			int ret = (cmdhandler[expected].func)
						(buffer, &conn_info);
			erase_str(buffer);
			switch (ret) {
				case CMD_HANDLED:
					/* expecting the next */
					expected++;
					break;
				case CMD_ERROR:
					/* reset counter, skip reset and
					 * quit function */
					ret = (cmdhandler[RESETFUNC].func)
							(buffer, &conn_info);
					expected = QUITFUNC + 1;
					/* errors are handled by the
					 * command handler functions
					 * */
					break;
			}
			/* found and called proper function */
		} else {
			sayf(clntinfo->clientsocket,
				"530 Login incorrect. Expected %scommand\r\n",
				cmdhandler[expected].cmd);
			/* reset counter, skip reset and quit function */
			(cmdhandler[RESETFUNC].func)
						(buffer, &conn_info);
			expected = QUITFUNC + 1;
		}
		if (clntinfo->login.stage == LOGIN_ST_FULL) {
			/* we are done */
			free(buffer); buffer = (char*) 0;
			return 0;
		}
	}
	/* li.host and li.user  are freed at the termination of the
	 * programm */
}

int login(struct clientinfo* clntinfo, int stage) {
	int ret = 0;

	if (! clntinfo->user || ! clntinfo->destination) {
		say(clntinfo->clientsocket,
				"500 Error logging in\r\n");
		log(4, "No user or destination was set - Cannot proceed");
		return -1;
	}

	if ((ret = login_setforward(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	if (stage >= LOGIN_ST_CONNECTED) {
		ret = login_connect(clntinfo);
		if (ret) { return ret; }
	}
	if (stage >= LOGIN_ST_USER) {
		ret = login_sendauth_user(clntinfo);
		if (ret) { return ret; }
	}
	if (stage >= LOGIN_ST_FULL) {
		ret = login_connect(clntinfo);
		if (ret) { return ret; }
		ret = login_auth(clntinfo);
		if (ret) {
			login_failed(clntinfo);
			return ret;
		} else {
			config_destroy_sectionconfig();
		}
	}
	return 0;
}


static
int login_connect(struct clientinfo* clntinfo) {
	int ret;

	if (clntinfo->login.stage >= LOGIN_ST_CONNECTED) {
		return 0;
	}

	if ((ret = login_init_connection(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	if ((ret = login_connected_setup(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	clntinfo->login.stage = LOGIN_ST_CONNECTED;
	return 0;
}


static
int login_auth(struct clientinfo* clntinfo) {
	int ret;

	if (clntinfo->login.stage < LOGIN_ST_USER) {
		if ((ret = login_sendauth_user(clntinfo)) < 0) {
			clntinfo->login.stage = LOGIN_ST_CONNECTED;
			return ret;
		}
		clntinfo->login.stage = LOGIN_ST_USER;
	}

	if ((ret = login_sendauth_pass(clntinfo)) < 0) {
		clntinfo->login.stage = LOGIN_ST_CONNECTED;
		return ret;
	}

	if ((ret = login_loggedin_setup(clntinfo)) < 0) {
		clntinfo->login.stage = LOGIN_ST_CONNECTED;
		return ret;
	}

	return 0;
}


static
int login_mayconnect(struct clientinfo* clntinfo) {
	int allowed = 0;
	unsigned long int target;
	unsigned long int host_ip;
	unsigned long int client_ip, server_ip;

	allowed = 0;

	/* we don't know yet if clntinfo->destination is a hostname like
	 * "somehost.foo.com" or if it is an IP in a char* like
	 * "212.117.232.20"
	 *
	 * So call inet_addr(), it should make the decision  :-)
	 *
	 * */
	target = inet_addr(clntinfo->destination);

	/* get the client_ip by asking information about the socket the
	 * client is connected */
	client_ip = get_uint_ip(GET_IP_CLIENT, clntinfo);
	server_ip = (unsigned long int) UINT_MAX;

	if (target == (unsigned long int) UINT_MAX) {
		/* clntinfo->destination may be a hostname, but it needn't.
		 * It may also be an invalid IP: "393.39.239.500" or
		 * anything else. Look it up.
		 * */
		host_ip = hostent_get_ip(&hostcache, clntinfo->destination);
		if (host_ip != (unsigned long int) UINT_MAX) {
			/* successful lookup */
			server_ip = host_ip;
		} else {
			/* it's not a valid IP and we could not look it up,
			 * so we could not get valid information about the
			 * destination host */
			log(7, "Nonsense destination (no IP and could not look up hostname): %s",
					clntinfo->destination);
			allowed = 0;
		}
	} else {
		/* okay, it's an IP */
		server_ip = inet_addr(clntinfo->destination);
	}

	if (server_ip != (unsigned long) UINT_MAX) {
		/* we could get a valid IP, now we can evaluate if client
		 * is allowed to connect to the destination host */
		log(9, "Checking all tags");
		config_shrink_config(client_ip,
				server_ip,
				clntinfo->destination,
				clntinfo->user,
				(char*) 0,
				clntinfo->destinationport,
				&hostcache,
				TAG_GLOBAL | TAG_FROM | TAG_TO | TAG_USER
				| TAG_PORT);
		allowed = strcmp(config_get_option("access"), "allow") == 0;
	}

	if (!allowed) {
		say(clntinfo->clientsocket, "531 You are not allowed to connect to that host.\r\n");
		log(8, "Not allowed to connect to %s", clntinfo->destination);
		li.respcode = 531;
		return -2;
	}
	return allowed;  /* a positive int */
}


static
void free_loginst(struct loginstruct* ls) {
	if (ls->fwpass) {
		free(ls->fwpass);
	}
	if (ls->login) {
		free(ls->login);
	}
	if (ls->pass) {
		free(ls->pass);
	}
	ls->fwpass = ls->login = ls->pass = 0;
}


static
int login_setforward(struct clientinfo* clntinfo) {

	const char* forward;
	struct loginstruct forward_info;
	struct slist_t* forward_list, *forward_list_cur;

	config_shrink_config(get_uint_ip(GET_IP_CLIENT, clntinfo),
			get_uint_ip(GET_IP_SERVER, clntinfo),
			clntinfo->destination,
			clntinfo->user,
			(char*) 0,
			clntinfo->destinationport,
			&hostcache,
			TAG_GLOBAL | TAG_FROM | TAG_TO | TAG_USER | TAG_PORT);

	if (strcmp(config_get_option("access"), "allow") != 0) {
		return -1;
	}

	/* see if there is an option for the forward */
	forward = config_get_option("forward");

	if (! forward) {
		/* no forward */
		li.userlogin = strdup(clntinfo->user);
		li.usereffective = strdup(clntinfo->user);
		li.userforwarded = strdup("<no forward>");
		enough_mem(li.userlogin);
		enough_mem(li.usereffective);
		enough_mem(li.userforwarded);
		return 0;
	}

	/*
	 * <user ftp>
	 *    forward johnfred@fooserver.com
	 *    forward johnfred@fooserver.com,3949,p
	 *    forward johnfred@fooserver.com     *     johnspass
	 *    forward johnfred@fooserver.com   JiKe94  johnspass
	 * </user>
	 */

	if ( ! (forward_list = config_split_line( forward, WHITESPACES )) ) {
		memset(&forward_info, 0, sizeof(struct loginstruct));
		return 0;
	}

	forward_list_cur = forward_list;

	forward_info.login = strdup(forward_list_cur->value);
	enough_mem(forward_info.login);

	forward_list_cur = forward_list_cur->next;

	if (forward_list_cur && forward_list_cur->value) {
		forward_info.fwpass = strdup(forward_list_cur->value);
		enough_mem(forward_info.fwpass);
	} else {
		forward_info.fwpass = (char*) 0;
	}

	if (forward_list_cur) {
		forward_list_cur = forward_list_cur->next;
	}
	if (forward_list_cur && forward_list_cur->value) {
		forward_info.pass = strdup(forward_list_cur->value);
		enough_mem(forward_info.pass);
	} else {
		forward_info.pass = (char*) 0;
	}
	slist_destroy(forward_list);

	/* check if the password matches */
	if (forward_info.fwpass) {
		if (strcmp(forward_info.fwpass, "*") == 0
			/* allow any pw but use destpw on dest */
			|| cryptcmp(forward_info.fwpass, clntinfo->pass)==0) {
			/* check for the exact password */

			clntinfo->pass = realloc(clntinfo->pass,
						strlen(forward_info.pass) + 1);
			enough_mem(clntinfo->pass);
			strncpy(clntinfo->pass, forward_info.pass,
				strlen(forward_info.pass) /* always true */);
		} else if (cryptcmp(forward_info.fwpass, clntinfo->pass) != 0){
			say(clntinfo->clientsocket, "530 Login incorrect\r\n");
			free_loginst(&forward_info);
			return -1;
		}
	}

	/* back up */
	clntinfo->before_forward.user = strdup(clntinfo->user);
	clntinfo->before_forward.destination = strdup(clntinfo->destination);
	clntinfo->before_forward.destinationport = clntinfo->destinationport;

	enough_mem(clntinfo->before_forward.user);
	enough_mem(clntinfo->before_forward.destination);

	/* delete those configuration values that are evaluated by the
	 * parsing routine  -  they are not necessary anymore */
	config_option_list_delete("defaultforward");
	config_option_list_delete("transparent-forward");

	/* call the parsing routine and let it set the values */
	if (set_userdest(forward_info.login, 0, clntinfo, "@,: \t") < 0) {
		return -1;
	}

	/* set values for log info struct */
	li.userlogin = strdup(clntinfo->before_forward.user);
	li.usereffective = strdup(clntinfo->user);
	li.userforwarded = strdup(forward_info.login);
	enough_mem(li.userlogin);
	enough_mem(li.usereffective);
	enough_mem(li.userforwarded);

	free_loginst(&forward_info);

	/* shrink the configuration again - with TAG_FORWARDED this time */

	config_shrink_config(get_uint_ip(GET_IP_CLIENT, clntinfo),
			get_uint_ip(GET_IP_SERVER, clntinfo),
			clntinfo->destination,
			clntinfo->user,
			clntinfo->before_forward.user,
			clntinfo->destinationport,
			&hostcache,
			TAG_GLOBAL | TAG_FROM | TAG_TO | TAG_USER
			| TAG_PORT | TAG_FORWARDED );

	return 0;
}


#define ERR_STR_P2 "Error connecting to %s: %s\r\n"

static
int login_init_connection(struct clientinfo* clntinfo) {
	int ss, cs = clntinfo->clientsocket;
	int ret, err;

	if ((ret = login_mayconnect(clntinfo)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	/* connect */
	ss = openportname(clntinfo->destination,
			  clntinfo->destinationport, 0, clntinfo);
	if (ss < 0) {
		err = errno;
		sayf(cs, "500 "ERR_STR_P2,
				clntinfo->destination, strerror(err));
		log(5, ERR_STR_P2, clntinfo->destination, strerror(err));
		return -1;
	}

	/* connected */
	clntinfo->serversocket = ss;

	return 0;
}
#undef ERRSTR


static
int login_connected_setup(struct clientinfo* clntinfo) {
	int ret;
	char* buffer;

	if ((ret = login_readwelcome(clntinfo)) < 0) {
		return ret;
	}

	if (clntinfo->transparent == TRANSPARENT_YES
		/* we are connected */
		&&
	    (config_get_option("logintime")
			&& strcasecmp(config_get_option("logintime"),
				"connect") == 0)) {
		say(clntinfo->clientsocket,clntinfo->login.welcomemsg.fullmsg);

		clntinfo->login.welcomemsg.fullmsg = (char*) 0;
		clntinfo->login.welcomemsg.lastmsg = (char*) 0;
	}

	say(clntinfo->serversocket, "SYST\r\n");
	buffer = ftp_readline(clntinfo->serversocket);
	if (!buffer) {
		if (timeout) {
			log(1, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(clntinfo->clientsocket);
			return -1;
		} else {
			err_readline(clntinfo->clientsocket);
			/* -2 ?? */
			return -1;
		}
	}
	if (!strlen(buffer)) {
		log(2, "The server did not respond to the initial SYST command");
		say(clntinfo->clientsocket,
				"The server did not respond correctly\r\n");
		free(buffer);
		return -1;
	}
	free(buffer);
	buffer = (char*) 0;


	/* get the throughput rate */
	clntinfo->throughput = config_get_foption("throughput", -1.0);

	clntinfo->addr_to_server =
			socketinfo_get_local_ip(clntinfo->serversocket);

	clntinfo->servermode = getservermode();

	return 0;
}


static
int login_readwelcome(struct clientinfo *clntinfo) {
	/* Read the welcome line */
	clntinfo->login.welcomemsg = readall(clntinfo->serversocket);
	if (!clntinfo->login.welcomemsg.fullmsg) {
		/* an error occurred */
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(clntinfo->clientsocket);
		} else {
			set_errstr("Server closed the connection");
			err_readline(clntinfo->clientsocket);
		}
		return -1;
	}

	log(9, "Connected to %s, got \"%s\" as welcome message",
		clntinfo->destination, clntinfo->login.welcomemsg.fullmsg);

	if (!checkbegin(clntinfo->login.welcomemsg.lastmsg, "220 ")) {
		log(2, "Not a valid FTP server response (%s)",
				clntinfo->login.welcomemsg.fullmsg);
		say(clntinfo->clientsocket,clntinfo->login.welcomemsg.fullmsg);

		free(clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg = (char*) 0;
		clntinfo->login.welcomemsg.lastmsg = (char*) 0;
		return -1;
	}
	return 0;
}


static
int login_sendauth_user(struct clientinfo* clntinfo) {
	size_t sendbufsize, ret;
	char* sendbuf;

	if (clntinfo->login.stage >= LOGIN_ST_USER) {
		return 0;
	}

	sendbufsize = strlen("USER \r\n") + strlen(clntinfo->user) + 1;
	sendbuf = (char*) malloc(sendbufsize);
	enough_mem(sendbuf);

	snprintf(sendbuf, sendbufsize, "USER %s\r\n", clntinfo->user);
	ret = say(clntinfo->serversocket, sendbuf);
	free(sendbuf);
	sendbuf = 0;
	if (ret < 0) {
		log(2, "Error writing the user name to the server: %s",
			strerror(errno));
		return -1;
	}

	clntinfo->login.authresp = readall(clntinfo->serversocket);

	if (!clntinfo->login.authresp.fullmsg) {
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
			err_time_readline(clntinfo->clientsocket);
		} else {
			err_readline(clntinfo->clientsocket);
		}
		free(clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg = (char*) 0;
		clntinfo->login.welcomemsg.lastmsg = (char*) 0;
		return -1;
	}

	clntinfo->login.stage = LOGIN_ST_USER;
	return 0;
}


static
int login_sendauth_pass(struct clientinfo* clntinfo) {
	size_t sendbufsize;
	int ret;
	char* sendbuf, *buffer;

	buffer = clntinfo->login.authresp.lastmsg;
	if ( !checkdigits(buffer, 331) && !checkdigits(buffer, 230)) {
		/* prepend the response from the server with an error code
		 * and return it to the client */
		sendbufsize = strlen(buffer) + 4 + 1;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		snprintf(sendbuf, sendbufsize, "500 %s", buffer);
		say(clntinfo->clientsocket, sendbuf);
		free(sendbuf);

		log(7, "Got \"%s\" after sending the username.", buffer);
		free(clntinfo->login.authresp.fullmsg);
		free(clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg = (char*) 0;
		clntinfo->login.welcomemsg.lastmsg = (char*) 0;
		clntinfo->login.authresp.fullmsg   = (char*) 0;
		clntinfo->login.authresp.lastmsg   = (char*) 0;
		return -1;
	}

	if (!checkdigits(buffer, 230)) {
		char* userdup;
		size_t size;

		sendbufsize = strlen("PASS \r\n") + strlen(clntinfo->pass) + 1;
		sendbuf = (char*) malloc(sendbufsize);
		enough_mem(sendbuf);
		mlock(sendbuf, sendbufsize);
		snprintf(sendbuf, sendbufsize, "PASS %s\r\n", clntinfo->pass);
		ret = say(clntinfo->serversocket, sendbuf);

		size = strlen(clntinfo->user) + 3;
		userdup = (char*) malloc( size );
		snprintf(userdup, size, " %s ", clntinfo->user);

		if (strstr(ANON_USERS, userdup)) {
			clntinfo->anon_user = clntinfo->pass;
			clntinfo->pass = (char*) 0;
			munlock(sendbuf, sendbufsize);
		} else {
			memset(sendbuf, (char) 0, sendbufsize);
			clntinfo->anon_user = strdup("UNKNOWN");
		}
		free(sendbuf);
		free(userdup);
		sendbuf = userdup = (char*) 0;
		free(clntinfo->login.authresp.fullmsg);
		clntinfo->login.authresp.fullmsg   = (char*) 0;
		clntinfo->login.authresp.lastmsg   = (char*) 0;

		if (ret < 0) {
			log(2, "Error writing the password to the server: %s",
				strerror(errno));
			free(clntinfo->login.welcomemsg.fullmsg);
			clntinfo->login.welcomemsg.fullmsg = (char*) 0;
			clntinfo->login.welcomemsg.lastmsg = (char*) 0;
			return -1;
		}
	}
	return 0;
}


static
int login_successfulp(struct clientinfo* clntinfo) {
	char* buffer;

	clntinfo->login.authresp = readall(clntinfo->serversocket);
	if (!clntinfo->login.authresp.fullmsg) {
		if (timeout) {
			log(2, "Timeout in %s line %d\n", __FILE__, __LINE__);
			err_time_readline(clntinfo->clientsocket);
		}
		else {
			err_readline(clntinfo->clientsocket);
		}
		if (clntinfo->login.welcomemsg.fullmsg) {
			free(clntinfo->login.welcomemsg.fullmsg);
		}
		return -1;
	}

	if (clntinfo->login.welcomemsg.fullmsg) {
		/* we have not yet sent the welcome message to the client */
		buffer = merge_responses(clntinfo->login.welcomemsg.fullmsg,
				clntinfo->login.authresp.fullmsg);
		/* free the welcome message */
		free(clntinfo->login.welcomemsg.fullmsg);
		clntinfo->login.welcomemsg.fullmsg
			= clntinfo->login.welcomemsg.lastmsg = (char*) 0;
	} else {
		/* if the welcome message is sent, just copy the
		 * authentication response */
		buffer = strdup(clntinfo->login.authresp.fullmsg);
		enough_mem(buffer);
	}

	li.respcode = getcode(clntinfo->login.authresp.lastmsg);

	if (!checkdigits(clntinfo->login.authresp.lastmsg, 230)) {
		say(clntinfo->clientsocket, clntinfo->login.authresp.lastmsg);
		log(8, "Got \"%s\" after sending the password",
				clntinfo->login.authresp.fullmsg);
		free(clntinfo->login.authresp.fullmsg);
		free(buffer);
		clntinfo->login.authresp.fullmsg   = (char*) 0;
		clntinfo->login.authresp.lastmsg   = (char*) 0;
		return -1;
	}
	free(clntinfo->login.authresp.fullmsg);
	clntinfo->login.authresp.fullmsg   = (char*) 0;
	clntinfo->login.authresp.lastmsg   = (char*) 0;

	/* Now the client receives the welcome message as well as the
	 * response of the server from the authentication with the same code
	 * at the beginning */
	say(clntinfo->clientsocket, buffer);
	free(buffer);
	buffer =0;

	/* successful login */
	log(6, "Logged in to %s as %s!",
		clntinfo->destination, clntinfo->user);

	clntinfo->login.stage = LOGIN_ST_FULL;
	return 0;
}


static
int login_loggedin_setup(struct clientinfo* clntinfo) {

	/* initialize the loginfo structure with the values that have to be
	 * set only once after a successful login */
	li.anon_user = clntinfo->anon_user;
	li.svrip = strdup(get_char_ip(GET_IP_SERVER, clntinfo));
	li.svrname = hostent_get_name(&hostcache, inet_addr(li.svrip));
	li.ifipsvr = strdup(conv_ip_to_char(clntinfo->addr_to_server));
	enough_mem(li.ifipsvr);
	li.svrlogin = clntinfo->destination;

	return login_successfulp(clntinfo);
}


static
int login_failed(struct clientinfo* clntinfo) {

	/* if there was a forward, restore the clntinfo fields from the
	 * before_forward values */

	if (clntinfo->before_forward.user) {
		free(clntinfo->user);
		clntinfo->user = clntinfo->before_forward.user;
		clntinfo->before_forward.user = (char*) 0;
	}
	if (clntinfo->before_forward.destination) {
		free(clntinfo->destination);
		clntinfo->destination = clntinfo->before_forward.destination;
		clntinfo->before_forward.destination = (char*) 0;

		clntinfo->destinationport
				= clntinfo->before_forward.destinationport;
	}

	/* switch configuration - remove the shrinked one,
	 * set the backup into place. If the login failed,
	 * the backup is used */

	config_activate_backup();


	/* shrink the configuration again - this will update the option list */

	config_shrink_config(get_uint_ip(GET_IP_CLIENT, clntinfo),
			(unsigned long int) UINT_MAX,
			(char*) 0,
			(char*) 0,
			(char*) 0,
			0,
			&hostcache,
			TAG_GLOBAL | TAG_FROM );

	return 0;
}



/* login handles the login to the server (the data is passed via CLNTINFO)
 *
 * Parameters: clntinfo: connection information
 * 
 * Return values: -1 for a system call failure (so that it can be
 *                   written via strerror to the client
 *                -2 for other errors
 *                 0 on success
 * Called by: handlecmds() after that version has read the username and
 *            maybe the password
 *
 *
 * Login response codes: 
 * 	Success:
 *		- forwards server response
 * 	Error:
 * 		- 530: Login incorrect
 * 		- 531: You are not allowed to connect to that host
 * 		- 532: Could not look up the destination host
 *		- 533: Error connecting to the server via ssh
 */





/*

	switch(clntinfo->servermode) {
#ifdef SFTP_SUPPORT
		case SECURE_SFTP:
			ret = sftp_login(clntinfo);
			break;
#endif
		default:
			ret = std_login(clntinfo);
	}

	if (ret) {
		return ret;
	}
*/




#ifdef SFTP_SUPPORT

static
int sftp_login(struct clientinfo* clntinfo) {

	int ret;
	char* args[6] = { "progname", "-l", clntinfo->user, "", "", "" };
	const char* crypt_pw;
	const char* ssh_options;
	int pw_match = 0;
	int nargs;
	struct in_addr in;

	if (!gethostaddr(clntinfo->destination,
				&in)) {
		say(clntinfo->clientsocket, "532 Could not look up the "
				"destination host\r\n");
		li.respcode = 532;
		return -2;
	}
	crypt_pw = get_sshpasswd(in.s_addr,
			clntinfo->user);
	if ( ! crypt_pw) {
		pw_match = 0;
		log(3, "No password specified for user %s",
				clntinfo->user);
	} else {
		pw_match = (cryptcmp(crypt_pw,
			clntinfo->pass) == 0);
	}
	if (!pw_match) {
		say(clntinfo->clientsocket, "530 Login incorrect.\r\n");
		li.respcode = 530;
		free(clntinfo->user);
		clntinfo->user =0;
		return -2;
	}

	ssh_options = get_sshoptions(in.s_addr, clntinfo->user);
	if (ssh_options) {
		args[3] = "-s";
		args[4] = strdup(ssh_options);
		args[5] = clntinfo->destination;
		enough_mem(args[5]);
		nargs = 6;
	} else {
		args[3] = clntinfo->destination;
		nargs = 4;
	}

	ret = sftp_open(args, nargs);
	if (nargs == 6) {
		free(args[4]);
	}
	if (ret) {
		log(2, "Error connecting to the sftp server");
		say(clntinfo->clientsocket, "533 Error connecting to the server via ssh\r\n"); 
		li.respcode = 533;
		return -2;
	}
	return 0;
}

#endif /* SFTP_SUPPORT */

